# Dev information

[Manual main page](homog2d_manual.md)

This page will gather misc. information useful for anyone wanting to contribute to code.

## Introduction

Git branches:

`master` branch is supposed to stay stable, with all tests passing.
`devX` branches are where things happen.

Building tests, demos, showcases, ...: everything is handled through the makefile, using the "out-of-source" principle:
everything that gets build ends up in the `BUILD` folder

## Makefile usage

### Available targets

* (default): same as `test`
* `clean`: erase everything that was built (in the `BUILD` folder)
* `test`: builds and runs the unit test with and without the `HOMOG2D_OPTIMIZE_SPEED` build option
* `testall`: builds and runs the unit test for the 3 different numerical types
* `check`: runs cppcheck (static analysis) ( https://cppcheck.sourceforge.io/ )
* `install`: `cp homog2d.hpp /usr/local/include`
* `doc`: build html reference (requires doxygen), with only public user API
* `doc-dev`: build full html reference (requires doxygen), including private class members
* `nobuild`: checks that the files in `misc/no_build` contain illegal code (part of the test process)


**Targets only available if Opencv is installed:**

* `demo`: build and run the Opencv interactive demo
* `showcase`: build and run the gif figures used in manual
* `test_fig`: build and run the code that generates the figures used in test cases
* `doc_fig`: build and run the code used to produce the figures of the manual

**Targets only available if LaTeX installed:**

* `doc_fig_tex`: build the LaTeX figures of the manual (TikZ based)

### Available options

These can be passed to make:
```
$ make <target> <option=Y|N>
```

* `USE_OPENCV`: enables the OpenCv additional features (useful for "test" targets)
* `USE_TINYXML2`: enables the SVG import (through Tinyxml2) additional features (useful for "test" targets)
* `USE_EIGEN`: enables the Eigen3 additional features
* `DEBUG`:  adds `-g` flag to compiler options

### Additional details

The `test_fig` target generates the images of some of the code used in the test cases.
This is solely to be able to see what is actually tested.
This code is located in `misc/figures_test`.
A valid program generating the image is generated by pasting together the considered file and `misc/figures_test/t_header.cxx` and one of the footer files
`misc/figures_test/t_footer_*.cpp`.
The makefile then compiles it and runs it.

## Code details

To be able to templatize all the code on the root numerical data type (float, double, ...), we implement some trick.
As the `LPBase` class is already templatized on the type (`type::IsPoint` or `type::IsLine`),
it would require a partial template specialization to define the behavior of each member function (or free function),
depending on the basic type (Line or Point), and still templatize on the numerical type.
C++ does not allow this.

Thus, the trick here is to call in each function a "sub" private function (prefixed with `impl_`) that gets overloaded by the datatype (point or line).
To achieve this overloading, each of these functions receives as additional (dummy) argument an object of type `BaseHelper`, templated by the numerical type.
In the definition of the function, this additional argument value is ignored,
it is there just so that the compiler can select the correct overload.

The different implementations are written as two `impl_` private functions that are templated by the numerical data type.
If the situation only makes sense for one of the types (for example `getAngle()` cannot be considered for two points), then
the implementation of that type only holds a `static_assert`, so that can be catched at build time.




## Coding style

- TABS, not spaces (1 byte per level)
- types have first character uppercase, variables and functions are lowercase
- `camelCase` for identifiers
- class member variables are prefixed with an underscore (`_`)
- spaces after parenthesis (`if( someBool )`)
- private and protected member function are prefixed with `p_` (or `impl_` for tag dispatch implementation)
- all symbols start with `HOMOG2D_`, to avoid name collisions

